__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import sys
import logging
import importlib
import builtins
import time
import json
import socket
from ca_apm_agent.utils import json_helpers
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.trace_enablers.virtual_stack import VirtualStack
from ca_apm_agent.trace_enablers.context import Context
from ca_apm_agent.connections.data import Data
from ca_apm_agent.instruments import import_hook
from ca_apm_agent.patchers.constants import *
from ca_apm_agent.connections.constants import PYTHON_MODULE, PYTHON_RASA_INTENT

logger = logging.getLogger(LOGGER_NAME[0] + '.patchers.monkey')


class Patcher:
    def __init__(self, cfg, data_queue):
        self.probe_dict = cfg.pop()
        self.to_instrument = cfg
        self.instrumented = []
        self.data_q = data_queue

    def patch(self):
        # Patches methods based on data in the instrument config file, CA_APM_PYTHON_ICFG env variable
        if not self.to_instrument:
            return
        i = 0
        while i < len(self.to_instrument):
            conf_dict = self.to_instrument[i]
            if not conf_dict[PATCHED]:
                _module = sys.modules.get(conf_dict[MODULE], None)
                if not _module:
                    i += 1
                    continue
                _class = None
                if conf_dict[CLASS] != 'None':
                    _class = getattr(_module, conf_dict[CLASS], None)
                    if not _class:
                        i += 1
                        continue
                _method = None
                if _class:
                    _method = getattr(_class, conf_dict[METHOD], None)
                else:
                    _method = getattr(_module, conf_dict[METHOD], None)
                if not _method:
                    i += 1
                    continue

                logger.debug('Disengaging import hook to import probes')
                builtins.__import__ = import_hook._wily_import_orig__
                probe_list = []
                for k in conf_dict[PROBE_KEYS]:
                    if not self.probe_dict[k][LOADED]:
                        self.probe_dict[k][LOADED] = getattr(importlib.import_module(self.probe_dict[k][PROBE_MODULE]),
                                                             self.probe_dict[k][PROBE_CLASS],
                                                             False)
                        if not self.probe_dict[k][LOADED]:
                            logger.error(
                                'Could not find probe class or module using provided probe module %s and class %s',
                                self.probe_dict[k][PROBE_MODULE], self.probe_dict[k][PROBE_CLASS])
                        else:
                            probe_list.append(self.probe_dict[k][LOADED])
                    else:
                        probe_list.append(self.probe_dict[k][LOADED])
                builtins.__import__ = import_hook.wily_import_hook
                logger.debug('Engaging import hook after attempting to import probes')
                if not probe_list:
                    logger.error(
                        'Unable to find any probe in instrumentation directive %s \nTherefore, not instrumenting',
                        str(conf_dict))
                    i += 1
                    continue

                if _class:
                    patched = self.patch_method(_class, _method, probe_list)
                    setattr(_class, conf_dict[METHOD], patched)
                else:
                    patched = self.patch_method(_module, _method, probe_list)
                    setattr(_module, conf_dict[METHOD], patched)
                conf_dict[PATCHED] = True
                self.instrumented.append(conf_dict)
                self.to_instrument.pop(i)
                import_hook.patch_list.discard(conf_dict[MODULE])
                logger.info('Instrumentation directive %s succeeded', str(conf_dict))

    def patch_method(self, target_class, target_func, probe_list):
        for i in range(0, len(probe_list)):
            probe_list[i] = probe_list[i]()

        def function_template(*args, **kwargs):
            function_name = target_class.__name__ + '.' + target_func.__name__
            try:
                logger.debug('Start Trace %s', str(target_func.__name__))
                vs = VirtualStack()
                trace_id = str(vs.get_trace_id())
                seq = vs.get_seq()
                logger.debug('Start Trace %s', str(function_name))
                context = Context(seq=seq, function_name=function_name, trace_id=trace_id)
                context.push_target_params(*args, **kwargs)
                vs.push(context)
                # call probe's start
                for probe in probe_list:
                    if function_name == 'Interpreter.parse' :
                       timestamp = int(round(time.time() * 1000))
                       context.set_timestamp(timestamp)
                    else:
                       probe.start(context)
                       call_msg = Data.create_data_func_call_msg_with_context(context_obj=context)
                       self.data_q.put_nowait(call_msg)
            except Exception as e:
                logger.error('Error starting probes for %s ', function_name)
                logger.debug('Exception that was raised %s', str(e))
            finally:
                # Call the method that was actually called
                try:
                    context.res = target_func(*args, **kwargs)
                except Exception:
                    context.e, context.v, context.tb = sys.exc_info()
                    context.func_name = function_name
            try:
                logger.debug('Finish Trace %s', str(target_func.__name__))
                # call probe's start
                for probe in probe_list:
                     if function_name == 'Interpreter.parse' :
                       probe.start(context)
                       call_msg = Data.create_data_func_call_msg_with_context(context_obj=context)
                       self.data_q.put_nowait(call_msg)  
                       
                for probe in reversed(probe_list):
                    probe.finish(context)
                vs.pop()
                ret_msg = Data.create_data_func_return_msg_with_context(context_obj=context)
                self.data_q.put_nowait(ret_msg)
            except Exception as e:
                logger.error('Error finishing probes for %s ', function_name)
                logger.debug('Exception that was raised %s', str(e))
            finally:
                if hasattr(context, 'e'):
                    raise context.e(context.v).with_traceback(context.tb)
                else:
                    return context.res

        return function_template



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
from ca_apm_agent.global_constants import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME[0] + '.instruments.import_hook')

_wily_import_orig__ = None
patch = None
patch_list = None
mod_set = set()
mod_len = len(mod_set)


def wily_import_hook(*args, **kwargs):
    if _wily_import_orig__ is not None:
        imported_module = _wily_import_orig__(*args, **kwargs)
        mod_set.add(imported_module)
        global mod_len
        if len(mod_set) != mod_len:
            mod_len = len(mod_set)
            logger.info('Module being loaded: %s', str(imported_module).strip('<>'))
            if imported_module.__name__ in patch_list:
                try:
                    if patch:
                        patch()
                    else:
                        logger.critical('Agent not started yet')
                except Exception as e:
                    logger.critical('FAILED TO INSTRUMENT :' + imported_module.__name__ + '\n' + str(e))
        return imported_module
    else:
        logger.debug('Import hook not initialized')



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.trace_enablers.virtual_stack import VirtualStack
from ca_apm_agent.correlation import correlation


def get_corid():
    vs = VirtualStack()
    if vs.size() > 1:
        ctx = vs.peekIndex(-2)
        if ctx.cor_id:
            http_corid = ctx.cor_id
        else:
            http_corid = correlation.get_new_corid_header()
    else:
        http_corid = correlation.get_new_corid_header()
    return http_corid


__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
import threading
import logging
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.connections.data import Data
from ca_apm_agent.utils import network

logger = logging.getLogger(LOGGER_NAME[0] + '.instruments.fork_handler')


def patch_method(target_func, host, port, instid, data_q, sleep_time):
    def function_template(*args, **kwargs):
        res = target_func(*args, **kwargs)
        try:
            if res == 0:
                # Change the logging file name
                tlogger = logging.getLogger(LOGGER_NAME[0])
                file_handler = None
                for x in tlogger.handlers:
                    if type(x).__name__ == 'RotatingFileHandler':
                        file_handler = x
                        break
                if file_handler:
                    t = file_handler.baseFilename.split('.')
                    processid = os.getpid()
                    containerid = network.getcontainerid()
                    if containerid:
                       processid = containerid
                    t[-2] = str(processid)
                    file_handler.baseFilename = '.'.join(t)
                    # Force a rollover so that the new logfile is created
                    file_handler.doRollover()
                    logger.debug('Changed logfile name')

                logger.info('Starting data connection for new process')
                data_conn = Data(host=host, port=port, instid=instid, data_msg_queue=data_q)
                data = threading.Thread(name='data_connection_worker', target=data_conn.worker,
                                        args=(sleep_time,))
                data.daemon = True
                data.start()
        except Exception as e:
            logger.error('Exception executing handler')
            logging.debug('Exception that was raised %s', str(e))
        finally:
            return res

    return function_template



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
PRINT_PREFIX = 'CA APM PYTHON AGENT: '
# Python Stuff
PYTHONPATH = 'PYTHONPATH'
PYTHONUNBUFFERED = 'PYTHONUNBUFFERED'
PYTHONVERBOSE = 'PYTHONVERBOSE'

# Agent stuff
LOGGER_NAME = []
CA_APM_PYTHON_ACFG = 'CA_APM_PYTHON_ACFG'  # Path to Agent Config File
CA_APM_PYTHON_ICFG = 'CA_APM_PYTHON_ICFG'  # Path to Instrumentation Config File
ACFG_MTIME = 'acfg_mtime'  # ACFG file modification time
ICFG_MTIME = 'icfg_mtime'  # ICFG file modification time
CA_APM_PYTHON_HOST_NAME = 'host'         # Hostname to connect to Collector Agent/ APMIA Package
CA_APM_PYTHON_PORT_NUM = 'port'          # Port number required to connect to Collector Agent/ APMIA Package
CA_APM_PYTHON_APP_NAME = 'app_name'      # Application Name to be displayed in EM



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
import subprocess
from ca_apm_agent.global_constants import LOGGER_NAME
import ca_apm_agent.status
from ca_apm_agent.connections._connect import Connect
from ca_apm_agent.utils import network

logger = logging.getLogger(LOGGER_NAME[0] + '.connections.command')


class Command:
    def __init__(self, host, port, instid, app_name, com_msg_queue):
        self.host = host
        self.port = port
        self.instid = instid
        self.app_name = app_name
        self.conn = Connect(host=self.host, port=self.port, logger=logger)
        self.conn_status = False
        self.msg_queue = com_msg_queue  # Thread Safe - make sure to check this for multiprocessing
        self.keep_alive = None
        self.collector_version = None
        self.params = None
        self.container_id = network.getcontainerid()

    def connect(self):
        for i in range(3):
            if self.conn.connect():
                self.conn_status = True
                self.send_msg(self._create_conn_msg())
                ca_apm_agent.status.com_ready[0] = True
                ca_apm_agent.status.com_restart[0] = False
                ca_apm_agent.status.com_pause[0] = False
                break

    def send_msg(self, msg):
        if self.conn_status:
            if not self.conn.send(msg):
                self.conn_status = False
                ca_apm_agent.status.com_ready[0] = False

    def recv_msg(self):
        if not self.conn_status:
            return False
        else:
            return self.conn.receive()

    @staticmethod
    def create_msg(list_of_tuples):
        from ca_apm_agent.utils import json_helpers
        return json_helpers.to_json_string(list_of_tuples)

    def _create_conn_msg(self):
        from ca_apm_agent.connections.constants import OP, COMMAND_CONN, PROBE, PYTHON, VER, ARF_VER, INSTID, PGM, PRMS
        from ca_apm_agent.utils import json_helpers

        l = [(OP, COMMAND_CONN), (PROBE, PYTHON), (VER, ARF_VER), (INSTID, self.instid), (PGM, self.app_name)]
        if self.container_id:
            from ca_apm_agent.connections.constants import CONTAINER_ID
            if self.params is None:
                self.params = {}
            self.params[CONTAINER_ID] = self.container_id

        if self.params is not None:
            l.append((PRMS, self.params))

        msg = json_helpers.to_json_string(l)
        return msg

    def _create_keep_alive_msg(self):
        if self.keep_alive is None:
            from ca_apm_agent.connections.constants import OP, ARF
            from ca_apm_agent.utils import json_helpers
            l = [(OP, ARF)]
            self.keep_alive = json_helpers.to_json_string(l)
        return self.keep_alive

    def worker(self):
        logger.info('Worker started')
        import time
        from ca_apm_agent.utils import json_helpers
        from ca_apm_agent.connections.constants import OP, CONFIG, COLLECTOR, VERSION, MESSAGE, SPEAK, PAUSE, CONTINUE, CMD, PRMS
        self.connect()
        while True:
            # Data connection sets this when its broken
            if ca_apm_agent.status.com_restart[0]:
                self.conn_status = False

            if self.conn_status:
                if not self.msg_queue.empty():
                    self.send_msg(self.msg_queue.get_nowait())
                recv_string = self.recv_msg()
                if recv_string:
                    try:
                        msgs = json_helpers.from_json_string(recv_string)
                    except Exception as e:
                        logger.error('Error Parsing  message from collector')
                        logger.debug('Message =' + recv_string)
                        logger.debug(str(e))
                    else:
                        for msg in msgs:
                            if msg[OP] == CONFIG:
                                if MESSAGE in msg:
                                    if COLLECTOR in msg[MESSAGE]:
                                        if VERSION in msg[MESSAGE][COLLECTOR]:
                                            self.collector_version = str(msg[MESSAGE][COLLECTOR][VERSION])
                                            logger.info('Connected to Collector %s', self.collector_version)
                                elif CMD in msg and msg[CMD] == VERSION:
                                    self.collector_version = str(msg[PRMS][COLLECTOR])
                                    logger.info('Connected to Collector %s', self.collector_version)
                                else:
                                    pass
                            elif msg[OP] == SPEAK:
                                self.send_msg(self._create_keep_alive_msg())
                                logger.debug('Received Keep-Alive, Sent Alive')
                            elif msg[OP] == PAUSE:
                                ca_apm_agent.status.com_pause[0] = True
                                logger.info('Received Pause message')
                            elif msg[OP] == CONTINUE:
                                ca_apm_agent.status.com_pause[0] = False
                                logger.info('Received Continue message')
                            else:
                                pass
                else:
                    if type(recv_string) == bool:
                        # The connection was broken on the other side and we will attempt to re-establish
                        logger.info('Will attempt to re-establish connection')
                        ca_apm_agent.status.com_restart[0] = True
                        ca_apm_agent.status.com_ready[0] = False
                    else:
                        # There was no data to read and receive passed an empty string
                        pass
            else:
                self.msg_queue.queue.clear()
                self.conn.close()
                self.connect()
                time.sleep(7)
                continue
            if self.msg_queue.empty():
                time.sleep(30)



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
from collections import OrderedDict
import json


def to_json_string(list_of_tuples):
    msg = json.dumps(OrderedDict(list_of_tuples))
    msg = msg.replace(', ', ',')
    msg = msg.replace(': ', ':')
    return msg


def from_json_string(string):
    string = string.replace('\r', '')
    string = string.replace('\n', '')
    string = string.replace('"{', '{')
    string = string.replace('}"', '}')
    string = string.replace('\\"', '"')
    string = string.split('}{')
    if len(string) > 1:
        string[0] = string[0] + '}'
        msg = [json.loads(string[0])]
        for i in range(1, len(string) - 1):
            string[i] = '{' + string[i] + '}'
            temp = json.loads(string[i])
            if temp != msg[-1]:
                msg.append(temp)
        string[-1] = '{' + string[-1]
        temp = json.loads(string[-1])
        if temp != msg[-1]:
            msg.append(temp)
        return msg
    msg = [json.loads(string[0])]
    return msg



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import socket
import subprocess


def getfqdn():
    return socket.getaddrinfo(socket.gethostname(), 0, 0, 0, 0, socket.AI_CANONNAME)[0][3]

def getcontainerid():
    temp = None
    containerid = None
    temp = subprocess.check_output(
            "grep docker -m1 /proc/self/cgroup|echo $(read s;s=${s##*/};s=${s#*docker-};s=${s%.scope};echo $s)",
            shell=True)
    containerid = temp.decode()
    if containerid == '\n':
       containerid = None
    else:
       containerid = containerid.strip('\n')
       
    return containerid



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
import os
import sys
import threading

from ca_apm_agent.config import read_instrument_cfg, read_agent_cfg, configure_logger, agent_config_monitor, instrument_config_monitor
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.connections.command import Command
from ca_apm_agent.connections.data import Data
from ca_apm_agent.patchers import fork_handler
from ca_apm_agent.patchers.monkey import Patcher
from ca_apm_agent.utils import network

logger = None


class Agent:
    patch = None

    def __init__(self, acfg, icfg, comm_q, data_q):
        # Below function sets attributes mentioned in CA_APM_PYTHON_ACFG
        self.apply_config(acfg)
        # After this is everything that are properties of the core agent that is static
        #  OR it cannot come through the agent config file
        processid = os.getpid()
        containerid = network.getcontainerid()
        
        if containerid:
           processid = containerid
           
        self.instid = processid
        self.icfg_mtime = icfg.pop()
        self.comm_q = comm_q
        self.data_q = data_q

        configure_logger(self.log, str(self.instid))

        # Logging can begin after this point.
        global logger
        logger = logging.getLogger(LOGGER_NAME[0] + '.agent')

        # After this all object instantiations
        self.comm_conn = Command(host=self.host, port=self.port, instid=self.instid, app_name=self.app_name,
                                 com_msg_queue=self.comm_q)
        self.data_conn = Data(host=self.host, port=self.port, instid=self.instid, data_msg_queue=self.data_q)
        self.patcher = Patcher(icfg, self.data_q)
        Agent.patch = self.patcher.patch

        # Logic to patch os.fork() - if it exists
        _os_ref = sys.modules['os']
        _fork_ref = getattr(_os_ref, 'fork', None)
        if _fork_ref:
            patched = fork_handler.patch_method(_fork_ref, self.host, self.port, self.instid, self.data_q,
                                                self.sleep_time)
            setattr(_os_ref, 'fork', patched)

        logger.debug('Agent attributes after creation %s', str(self.__dict__))

    def apply_config(self, cfg):
        for k, v in cfg.items():
            setattr(self, k, v)

    def start_agent(self):
        agent_config_monitor(self)
        instrument_config_monitor()
        comm = threading.Thread(name='command_connection_worker', target=self.comm_conn.worker)
        data = threading.Thread(name='data_connection_worker', target=self.data_conn.worker, args=(self.sleep_time,))
        comm.daemon = True
        data.daemon = True
        try:
            comm.start()
            data.start()
        except Exception as e:
            logger.error('Exception starting workers')
            logger.debug('Exception that was raised %s', str(e))



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.python_modifiers.singleton import Singleton
from ca_apm_agent.trace_enablers.virtual_stack import VirtualStack
from ca_apm_agent.connections.constants import TRACE_ID, CORR_ID
from ca_apm_agent.correlation import correlation
from ca_apm_agent.utils import corrid_helpers

try:
    import urllib.parse as parse
except ImportError:
    import urlparse as parse

logger = logging.getLogger(LOGGER_NAME[0] + '.probes.urllib.http')
# Constants
URL = 'url'
HOST_NAME = 'hostName'
HOST_PORT = 'hostPort'

class HTTPUrllibProbe(Singleton):
    def __init__(self):
        pass

    def start(self, context):
        logger.debug('Function Arguments: %s **** %s', str(context.args), str(context.kwargs))
        req = context.args[2]
        url = req.full_url
        host_info = parse.urlparse(req.full_url)
        host_name = host_info.hostname
        host_port = host_info.port 

        corId = corrid_helpers.get_corid()
        traceId = correlation.get_txn_traceid(corId)
        correlationId = {'CORID':corId}
        req.unredirected_hdrs.update(correlationId)
        context.cor_id = corId
        http_corid = corId
        
        context.set_params({URL: url, HOST_NAME: host_name, HOST_PORT: host_port, TRACE_ID: traceId, CORR_ID: http_corid})
        logger.debug('Pushed params: %s', str(context.params))           

    def finish(self, context):
        pass


__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
import logging
import logging.config
from ca_apm_agent.global_constants import CA_APM_PYTHON_ACFG, ACFG_MTIME, PRINT_PREFIX, LOGGER_NAME
from ca_apm_agent.utils import network

logger = None


def read_instrument_cfg(cfg, fn):
    # Do not use logging here!
    from ca_apm_agent.patchers.constants import MODULE, CLASS, METHOD, PROBE_CLASS, PROBE_MODULE, PROBE_KEYS, PATCHED, LOADED

    ret_list = []
    probe_dict = {}
    patch_list = set()
    i = 1
    for line in cfg:
        if line.startswith('#') or line.startswith('\n'):
            i += 1
            continue
        line = line.replace(' ', '')

        if not line.startswith('('):
            try:
                key, values = line.split('=')
            except Exception as e:
                print (PRINT_PREFIX + fn + ':Something wrong in line ' + str(i) + '. Ignoring line and continuing')
                i += 1
                continue
            else:
                values = values.strip('()\r\n')
                v1, v2 = values.split(',')
                probe_dict[key] = {PROBE_MODULE: v1, PROBE_CLASS: v2, LOADED: False}

        else:
            line = line.strip('()\r\n')
            words = line.split(',')
            flag = False
            try:
                for key in words[3:]:
                    if key not in probe_dict:
                        print (PRINT_PREFIX + fn + 'Probe variable ' + key + ' not declared. Ignoring line ' + str(
                            i) + ' and continuing')
                        flag = True
                        break
                if flag:
                    i += 1
                    continue
                ret_list.append(
                    {MODULE: words[0],
                     CLASS: words[1],
                     METHOD: words[2],
                     PROBE_KEYS: words[3:],
                     PATCHED: False})
                patch_list.add(words[0])
            except:
                print (PRINT_PREFIX + fn + ':Something wrong in line ' + str(i) + '. Ignoring line and continuing')
                i += 1
                continue
        i += 1
    if ret_list:
        return ret_list, probe_dict, patch_list
    else:
        return None


def read_agent_cfg(cfg, fn):
    # Do not use logging here!
    import ast

    ret_dict = {}
    i = 1
    for line in cfg:
        t_dict = ret_dict
        if line.startswith('#') or line.startswith('\n'):
            i += 1
            continue
        try:
            keys, value_ = line.strip('\n').split('=', 1)
        except Exception as e:
            print (PRINT_PREFIX + fn + ':Something wrong in line ' + str(i) + '. Ignoring line and continuing')
            i += 1
            continue
        keylist = keys.split('.')
        for k in keylist[:-1]:
            if k not in t_dict:
                try:
                    t_dict[k] = {}
                except TypeError:
                    print (PRINT_PREFIX + fn + ':You are trying to nest a key ' + str(
                        keylist) + ' who has already been assigned a constant value')
                    return None
            t_dict = t_dict[k]
        try:
            t_dict[keylist[-1]] = ast.literal_eval(value_)
        except ValueError:
            t_dict[keylist[-1]] = value_
        except TypeError:
            print (PRINT_PREFIX + fn + ':You are trying to nest a key ' + str(
                keylist) + ' who has already been assigned a constant value')
            return None
        i += 1

    if ret_dict:
        return ret_dict
    else:
        return None


def configure_logger(cfg_dict, mpid):
    logging.config.dictConfig(cfg_dict)
    global logger
    logger = logging.getLogger(LOGGER_NAME[0] + '.config')
    # Changing log filename to include process id
    file_handler = None
    for x in logger.parent.handlers:
        if type(x).__name__ == 'RotatingFileHandler':
            file_handler = x
            break
    t = file_handler.baseFilename.split('.')
    if len(t) > 1:
        t.insert(-1, mpid)
        if os.path.exists(file_handler.baseFilename):
            # Initial file without PID being deleted
            os.remove(file_handler.baseFilename)
        file_handler.baseFilename = '.'.join(t)
    else:
        t.append(mpid)
        t.append('log')
        if os.path.exists(file_handler.baseFilename):
            # Initial file without PID being deleted
            os.remove(file_handler.baseFilename)
        file_handler.baseFilename = '.'.join(t)
    # Force a rollover so that the new logfile is created
    file_handler.doRollover()
    logger.debug('Logger configured and ready to work')


def agent_config_monitor(agent_object):
    import threading
    import os
    import time
    logger.debug('Attempting to start monitor')

    def aconfig_worker(agent_object):
        logger.info('Monitor started')
        while True:
            if agent_object.acfg_mtime != os.path.getmtime(os.environ[CA_APM_PYTHON_ACFG]):
                logger.info('Config file change detected')
                with open(os.environ[CA_APM_PYTHON_ACFG], 'r') as f:
                    cfg = f.readlines()
                cfg = read_agent_cfg(cfg, os.environ[CA_APM_PYTHON_ACFG])
                if cfg is None:
                    logger.warning('CFG file has some issues (check logs), please correct it')
                else:
                    cfg[ACFG_MTIME] = os.path.getmtime(os.environ[CA_APM_PYTHON_ACFG])
                    agent_object.apply_config(cfg)
                    processid = os.getpid()
                    containerid = network.getcontainerid()
                    
                    if containerid:
                       processid = containerid
                       
                    configure_logger(agent_object.log,str(processid))
                    logger.info('Changes applied')
            time.sleep(60)

    t = threading.Thread(name='agent_config_monitor', target=aconfig_worker, args=(agent_object,))
    t.daemon = True
    t.start()


def instrument_config_monitor():
    # TODO
    # Dynamic instrumentation in multi processing... phewwww  READ SIGHUP :D HAHAHAHA
    pass



import logging
import base64
import string
import array
import uuid
import random
from ca_apm_agent.connections.constants import CORIDTEMPLATE
from ca_apm_agent.global_constants import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME[0] + '.correlation.correlation')

def get_new_corid_header():

	try:
		cor_len = 32
		corlist = [random.choice(string.digits+string.ascii_letters) for i in range(cor_len)]
		corid = ("".join(corlist)).upper()
		logger.debug('Random CorId Generated %s', corid)
	except Exception as e:
		logger.error('Problem In Generating Random Corid')
    
	encode_string = CORIDTEMPLATE
    
	decode_str = base64.b64decode(encode_string)
	logger.debug('Successfully Decoded The String %s', decode_str)
    
	arrobj = array.array('B', decode_str)
	logger.debug('Converted the Decoded string to Array Object %s', arrobj)
    
	bytevalues = bytearray(arrobj)
	length = bytevalues [96] 
	KEY_LEN = length
	keylist = [random.choice(string.digits+string.ascii_letters) for i in range(KEY_LEN)]
	txnid = ("".join(keylist)).upper()
    
	decode_bytevals = bytevalues.decode('cp855')        
	str_rep = str(decode_bytevals) 
	stxnid = str(txnid)
	logger.debug('New Txn Trace Id generated Successfully %s', stxnid)
    
	repstr = str_rep.replace(str_rep[97:97+KEY_LEN], stxnid)
	logger.debug('New Txn Trace Id Replaced Successfully')
    
	enc_repstr = repstr.encode('cp855')      
	final = base64.b64encode(enc_repstr)
	final_str = str(final, 'utf-8')
    
	comma = ','
	seqId = '1:1'
	propflag = '0'
	seqflag = '0'
    
	header = corid + comma + seqId + comma + propflag + comma + seqflag + comma + comma + comma + final_str
	logger.debug('New Header Sent Successfully %s', header)
	return header


def upd_corid_header(header):

	http_header = header
	
	if http_header is None:
		logger.error('Header Received is NULL')
	else:
		logger.debug('Header Received Successfuly')
    
	corid_received = http_header[:32]
	seqid_separators = http_header[32:43]
	encode_string = http_header[43:]
    
	decode_str = base64.b64decode(encode_string)
	logger.debug('Successfully Decoded The String %s', decode_str)
     
	arrobj = array.array('B', decode_str)
	logger.debug('Converted the Decoded string to Array Object %s', arrobj)
    
	bytevalues = bytearray(arrobj)
	length = bytevalues [96] 
	KEY_LEN = length
	keylist = [random.choice(string.digits+string.ascii_letters) for i in range(KEY_LEN)]
	txnid = ("".join(keylist)).upper()
    
	decode_bytevals = bytevalues.decode('cp855')        
	str_rep = str(decode_bytevals) 
	stxnid = str(txnid)
	logger.debug('New Txn Trace Id generated Successfully %s', stxnid)
    
	repstr = str_rep.replace(str_rep[97:97+KEY_LEN], stxnid)
	logger.debug('New Txn Trace Id Replaced Successfully')	
    
	enc_repstr = repstr.encode('cp855')      
	final = base64.b64encode(enc_repstr)
	final_str = str(final, 'utf-8')
    
	http_header = corid_received + seqid_separators + final_str
	logger.debug('Updated the Header Successfully %s', http_header)
	return http_header
    
def get_txn_traceid(header):

	http_header = header

	if http_header is None:
		logger.error('Header Received is NULL')
	else:
		logger.debug('Header Received Successfuly')

	corid_received = http_header[:32]
	seqid_separators = http_header[32:43]
	encode_string = http_header[43:]

	decode_str = base64.b64decode(encode_string)
		 
	arrobj = array.array('B', decode_str)

	bytevalues = bytearray(arrobj)
	length = bytevalues [96] 
	KEY_LEN = length

	decode_bytevals = bytevalues.decode('cp855')        
	str_rep = str(decode_bytevals) 

	traceid = str_rep[97:97+KEY_LEN]
		
	logger.debug('Txn TraceID Sent Successfully %s', traceid)
	return traceid


__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
import time
import json
import socket
from ca_apm_agent.utils import json_helpers
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.python_modifiers.singleton import Singleton
from ca_apm_agent.connections.constants import PYTHON_MODULE, PYTHON_RASA_INTENT

logger = logging.getLogger(LOGGER_NAME[0] + '.probes.rasa.rasa')

class RasaProbe(Singleton):
    def __init__(self):
        pass

    def start(self, context):
        logger.debug('Rasa Probe -Start Method')
        context_res=context.res
        message_id = context.args[2]
        json_string = json_helpers.from_json_string(json.dumps(context.res))
        intent_name = json_string[0]['intent']['name']
        intent_confidence = json_string[0]['intent']['confidence']
        user_input = json_string[0]['text']
        hostName = socket.gethostname()
        port = "5005"
        context.set_params({"url": intent_name, "hostName": hostName, "hostPort": port, "Confidence": intent_confidence, "UserInput": user_input,
                            "MessageId": message_id, PYTHON_MODULE: "Rasa", PYTHON_RASA_INTENT: intent_name})


    def finish(self, context):
        logger.debug('Rasa Probe - Finish Method')



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
import sys
import builtins
from queue import Queue
import ca_apm_agent.config as config
from ca_apm_agent.global_constants import PRINT_PREFIX, LOGGER_NAME

environ = os.environ


def load_agent_config():
    from ca_apm_agent.global_constants import CA_APM_PYTHON_ACFG, ACFG_MTIME, CA_APM_PYTHON_HOST_NAME, CA_APM_PYTHON_PORT_NUM, CA_APM_PYTHON_APP_NAME
    flag = False
    cfg = None
    if CA_APM_PYTHON_ACFG in environ:
        try:
            # Loading config file
            cfg_file = open(environ[CA_APM_PYTHON_ACFG], 'r')
            cfg = cfg_file.readlines()
        except Exception as e:
            print (PRINT_PREFIX + 'Issues opening/reading provided agent config file, please check file and restart')
            print (str(e))
            print (PRINT_PREFIX + 'Using default agent config file')
            flag = True
        finally:
            cfg_file.close()
    else:
        flag = True

    if not flag:
        cfg = config.read_agent_cfg(cfg, environ[CA_APM_PYTHON_ACFG])
        if cfg is not None:
            cfg[ACFG_MTIME] = os.path.getmtime(environ[CA_APM_PYTHON_ACFG])
        else:
            print (PRINT_PREFIX + 'Provided Agent CFG file has some semantic issues - Please correct and restart')
            print (PRINT_PREFIX + 'Switching to default agent config file')
            flag = True

    if flag:
        import ca_apm_agent.directives as directives
        environ[CA_APM_PYTHON_ACFG] = os.path.join(os.path.dirname(os.path.realpath(directives.__file__)),
                                                   'agent.profile')
        try:
            cfg_file = open(environ[CA_APM_PYTHON_ACFG], 'r')
            cfg = cfg_file.readlines()
        except Exception as e:
            print (PRINT_PREFIX + 'Issues opening default agent config file. ' \
                                 'Possible corruption / incorrect installation of agent')
            print (str(e))
            print (PRINT_PREFIX + 'Will proceed without loading the agent')
            return None
        finally:
            cfg_file.close()
        cfg = config.read_agent_cfg(cfg, environ[CA_APM_PYTHON_ACFG])
        if cfg is not None:
            cfg[ACFG_MTIME] = os.path.getmtime(environ[CA_APM_PYTHON_ACFG])
        else:
            print (PRINT_PREFIX + 'Default Agent Config has semantic issues - Proceeding without loading the agent')

    if "CA_APM_PYTHON_HOST_NAME" in environ:
       cfg[CA_APM_PYTHON_HOST_NAME]=environ['CA_APM_PYTHON_HOST_NAME']
    else:
        if "ca_apm_python_host_name" in environ:
            cfg[CA_APM_PYTHON_HOST_NAME]=environ['ca_apm_python_host_name']

    if "CA_APM_PYTHON_PORT_NUM'" in environ:
       cfg[CA_APM_PYTHON_PORT_NUM]=environ['CA_APM_PYTHON_PORT_NUM']
    else:
        if "ca_apm_python_port_num" in environ:
            cfg[CA_APM_PYTHON_PORT_NUM]=environ['ca_apm_python_port_num']
 
    if "CA_APM_PYTHON_APP_NAME" in environ:
       cfg[CA_APM_PYTHON_APP_NAME]=environ['CA_APM_PYTHON_APP_NAME']
    else:
        if "ca_apm_python_app_name" in environ:
            cfg[CA_APM_PYTHON_APP_NAME]=environ['ca_apm_python_app_name']
    
    return cfg


def load_instrument_config():
    from ca_apm_agent.global_constants import CA_APM_PYTHON_ICFG, ICFG_MTIME
    flag = False
    cfg = None
    probe_dict = None
    patch_list = None
    if CA_APM_PYTHON_ICFG in environ:
        try:
            # Loading config file
            cfg_file = open(environ[CA_APM_PYTHON_ICFG], 'r')
            cfg = cfg_file.readlines()
        except Exception as e:
            print (PRINT_PREFIX + 'Issues opening/reading provided instrument config file, please check file and restart')
            print (str(e))
            print (PRINT_PREFIX + 'Using default agent config file')
            flag = True
        finally:
            cfg_file.close()
    else:
        flag = True

    if not flag:
        cfg, probe_dict, patch_list = config.read_instrument_cfg(cfg, environ[CA_APM_PYTHON_ICFG])
        if cfg is not None:
            cfg.append(os.path.getmtime(environ[CA_APM_PYTHON_ICFG]))
        else:
            print (PRINT_PREFIX + 'Provided Instrument CFG file has some semantic issues - Please correct and restart')
            print (PRINT_PREFIX + 'Switching to default agent config file')
            flag = True

    if flag:
        import ca_apm_agent.directives as directives
        environ[CA_APM_PYTHON_ICFG] = os.path.join(os.path.dirname(os.path.realpath(directives.__file__)),
                                                   'instrument.pbd')
        try:
            cfg_file = open(environ[CA_APM_PYTHON_ICFG], 'r')
            cfg = cfg_file.readlines()
        except Exception as e:
            print (PRINT_PREFIX + 'Issues opening default instrument config file. ' \
                                 'Possible corruption / incorrect installation of agent')
            print (str(e))
            print (PRINT_PREFIX + 'Will proceed without loading the agent')
            return None
        finally:
            cfg_file.close()
        cfg, probe_dict, patch_list = config.read_instrument_cfg(cfg, environ[CA_APM_PYTHON_ICFG])
        if cfg is not None:
            cfg.append(probe_dict)
            cfg.append(os.path.getmtime(environ[CA_APM_PYTHON_ICFG]))
        else:
            print (PRINT_PREFIX + 'Default Agent Config has semantic issues - Proceeding without loading the agent')

    return cfg, patch_list


if __name__ == 'sitecustomize':
    # Script starts here
    # Removing this file's path

    try:
        sys.path.remove(os.path.dirname(__file__))
    except ValueError:
        pass

    agent_cfg = load_agent_config()
    instrument_cfg, patch_list = load_instrument_config()
    LOGGER_NAME.extend(list(agent_cfg['log']['loggers']))  # list(dict) is a safe way to extract keys in py2&3

    if (agent_cfg is not None) and (instrument_cfg is not None):
        from ca_apm_agent.agent import Agent

        agent = Agent(agent_cfg, instrument_cfg, Queue(), Queue())

        import logging

        logger = logging.getLogger(LOGGER_NAME[0] + '.bootstrap.sitecustomize')

        try:
            agent.start_agent()
        except Exception as e:
            logger.error('Problems initializing agent')
            logger.debug('Problem encountered %s', str(e))
        else:
            # This import is critical and should be only done after the agent has started
            from ca_apm_agent.instruments import import_hook

            import_hook._wily_import_orig__ = builtins.__import__
            import_hook.patch = Agent.patch
            import_hook.patch_list = patch_list
            builtins.__import__ = import_hook.wily_import_hook
            logger.info('Import hook engaged')
        finally:
            pass

    logger.info('Attempting to remove injected sitecustomize')
    logger.debug('sys.modules %s', sys.modules)
    if sys.modules.pop('site', None) is not None:
        import logging
        import imp
        import sys
        from ca_apm_agent.global_constants import LOGGER_NAME

        logger = logging.getLogger(LOGGER_NAME[0] + '.bootstrap.sitecustomize')
        logger.info('Attempt Succeeded')
    else:
        logger.info('Attempt Failed - see debug logs')

    logger.debug('Attempting to load target app/environ sitecustomize')
    try:
        mod_data = imp.find_module('site')
        mod = imp.load_module('site', *mod_data)
        sys.modules['site'] = mod
        logger.debug('Attempt Succeeded')
    except ImportError:
        logger.warning('Unable to find target app/environ sitecustomize, if any')
    except Exception as e:
        logger.error('Exception in loading app/environ sitecustomize')
        logger.debug('Exception that was raised %s', str(e))



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import threading
from ca_apm_agent.python_modifiers.singleton import Singleton

JAVA_MAX = 9223372036854775807


class VirtualStack(Singleton):
    local_store = threading.local()

    def __init__(self):
        self.max = JAVA_MAX
        self.ls = VirtualStack.local_store

    def init(self):
        if not hasattr(self.ls, 'seed'):
            self.ls.seed = 0
            self.ls.seq = self.ls.seed
            self.ls.trace_id = self.ls.seed
            self.ls.stack = []

    def is_empty(self):
        if self.ls.stack:
            return False
        else:
            return True

    def push(self, value):
        self.ls.stack.append(value)

    def pop(self):
        if self.ls.stack:
            return self.ls.stack.pop()
        else:
            return None

    def peek(self):
        if self.ls.stack:
            return self.ls.stack[-1]
        else:
            return None
            
    def peekIndex(self, index):
        try:
            if self.ls.stack:
                return self.ls.stack[index]
            else:
                return None
        except IndexError:
            return None            

    def size(self):
        return len(self.ls.stack)

    def get_seq(self):
        self.ls.seq += 1
        if self.ls.seq > self.max:
            self.ls.seq = self.ls.seed + 1
        return self.ls.seq

    def get_trace_id(self):
        if not self.ls.stack:
            self.ls.trace_id += 1
            if self.ls.seq + 2000 > self.max:
                self.ls.seq = self.ls.seed
            if self.ls.trace_id > self.max:
                self.ls.trace_id = self.ls.seed + 1
        return self.ls.trace_id



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
from ca_apm_agent.python_modifiers.singleton import Singleton


# Only this Probe class should be called TProbe. No other class should be called TProbe
class TProbe(Singleton):
    def __init__(self):
        pass

    def start(self, context):
        pass

    def finish(self, context):
        pass



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
import time
import os
from ca_apm_agent.global_constants import LOGGER_NAME
import ca_apm_agent.status
from ca_apm_agent.connections._connect import Connect
from ca_apm_agent.utils import network

logger = logging.getLogger(LOGGER_NAME[0] + '.connections.data')


class Data:
    def __init__(self, host, port, instid, data_msg_queue):
        self.host = host
        self.port = port
        self.instid = instid
        processid = os.getpid()
        containerid = network.getcontainerid()
        
        if containerid:
           processid = containerid
           
        self.pid = processid
        self.conn = Connect(host=self.host, port=self.port, logger=logger)
        self.conn_status = False
        self.msg_queue = data_msg_queue

    def connect(self, delay=10):
        if delay:
            time.sleep(delay)
        while (not ca_apm_agent.status.com_ready[0]) and (self.pid == self.instid):
            logger.debug('Waiting for Command Connection to be established')
            time.sleep(20)
        for i in range(3):
            if self.conn.connect():
                self.conn_status = True
                self.send_msg(self._create_conn_msg())
                break

    def send_msg(self, msg):
        if (not ca_apm_agent.status.com_ready[0]) and (self.pid == self.instid):
            self.conn_status = False
        if self.conn_status and (not ca_apm_agent.status.com_pause[0]):
            if not self.conn.send(msg):
                self.conn_status = False
                processid = os.getpid()
                containerid = network.getcontainerid()
                if containerid:
                   processid = containerid
                if processid == self.instid:
                    ca_apm_agent.status.com_restart[0] = True
                    ca_apm_agent.status.com_ready[0] = False

    def recv_msg(self):
        if not self.conn_status:
            return False
        else:
            return self.conn.receive()

    @staticmethod
    def create_msg(list_of_tuples):
        from ca_apm_agent.utils import json_helpers
        return json_helpers.to_json_string(list_of_tuples)

    def _create_conn_msg(self):
        from ca_apm_agent.connections.constants import OP, DATA_CONN, PROBE, PYTHON, VER, ARF_VER, INSTID, PID
        from ca_apm_agent.utils import json_helpers

        l = [(OP, DATA_CONN), (PROBE, PYTHON), (VER, ARF_VER), (INSTID, self.instid), (PID, self.pid)]
        msg = json_helpers.to_json_string(l)
        return msg

    @staticmethod
    def create_data_func_call_msg(pid, tid, seq, func_name=None, func_id=None, params=None):
        from ca_apm_agent.connections.constants import OP, FUNCTION_CALL, FUNCTION_NAME, FUNCTION_ID, TIMESTAMP, PID, TID, PRMS, SEQ
        from ca_apm_agent.utils import json_helpers
        l = [(OP, FUNCTION_CALL)]
        if func_name is not None:
            l.append((FUNCTION_NAME, func_name))
        if func_id is not None:
            l.append((FUNCTION_ID, func_id))
        l.extend([(TIMESTAMP, int(round(time.time() * 1000))), (PID, pid), (TID, tid), (SEQ, seq)])
        if params is not None:
            l.append((PRMS, params))
        msg = json_helpers.to_json_string(l)
        return msg

    @staticmethod
    def create_data_func_call_msg_with_context(context_obj):
        from ca_apm_agent.connections.constants import OP, FUNCTION_CALL, FUNCTION_NAME, FUNCTION_ID, TIMESTAMP, PID, TID, PRMS, SEQ
        from ca_apm_agent.utils import json_helpers
        l = [(OP, FUNCTION_CALL)]
        if context_obj.func_name is not None:
            l.append((FUNCTION_NAME, context_obj.func_name))
        if context_obj.func_id is not None:
            l.append((FUNCTION_ID, context_obj.func_id))
        if context_obj.timestamp is not None:
           l.extend([(TIMESTAMP, context_obj.timestamp), (PID, context_obj.pid), (TID, context_obj.tid),
                     (SEQ, context_obj.seq)])
        else:
            l.extend([(TIMESTAMP, int(round(time.time() * 1000))), (PID, context_obj.pid), (TID, context_obj.tid),
                      (SEQ, context_obj.seq)])
        if context_obj.params is not None:
            l.append((PRMS, context_obj.params))
        msg = json_helpers.to_json_string(l)
        return msg

    @staticmethod
    def create_data_func_return_msg(pid, tid, cseq, exception_dict=None):
        from ca_apm_agent.connections.constants import OP, FUNCTION_RET, TIMESTAMP, PID, TID, CSEQ, EXCEPTION
        from ca_apm_agent.utils import json_helpers
        l = [(OP, FUNCTION_RET)]
        l.extend([(TIMESTAMP, int(round(time.time() * 1000))), (PID, pid), (TID, tid), (CSEQ, cseq)])
        if exception_dict is not None:
            l.append((EXCEPTION, exception_dict))
        msg = json_helpers.to_json_string(l)
        return msg

    @staticmethod
    def create_data_func_return_msg_with_context(context_obj):
        from ca_apm_agent.connections.constants import OP, FUNCTION_RET, TIMESTAMP, PID, TID, CSEQ, EXCEPTION, PRMS
        from ca_apm_agent.utils import json_helpers
        l = [(OP, FUNCTION_RET)]
        l.extend([(TIMESTAMP, int(round(time.time() * 1000))), (PID, context_obj.pid), (TID, context_obj.tid),
                  (CSEQ, context_obj.seq)])
        if context_obj.params is not None:
            l.append((PRMS, context_obj.params)) 
        if context_obj.exc is not None:
            l.append((EXCEPTION, context_obj.exc))
        msg = json_helpers.to_json_string(l)
        return msg

    def worker(self, sleep_time):
        logger.info('Worker started')
        import time
        self.connect()
        while True:
            if self.conn_status:
                while (not self.msg_queue.empty()) and (self.pid != self.instid or ca_apm_agent.status.com_ready[0]):
                    self.send_msg(self.msg_queue.get_nowait())
            else:
                self.conn.close()
                self.connect()
            if self.msg_queue.empty():
                time.sleep(sleep_time)



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
OP = 'op'
PROBE = 'probe'
COLLECTOR = 'collector'
VER = 'ver'
VERSION = 'version'
INSTID = 'instid'
PGM = 'pgm'
CMD = 'cmd'
MOD = 'mod'
PRMS = 'prms'
QUERY = 'query'
URL = 'url'
PID = 'pid'
TID = 'tid'
TIMESTAMP = 'ts'
SEQ = 'seq'
CSEQ = 'cseq'
FUNCTION_NAME = 'fn'
FUNCTION_ID = 'fnid'
METRIC_ID = 'mid'
METRIC_TYPE = 'mtype'
VALUE = 'val'
COUNT = 'ct'
MESSAGE = 'msg'
EXCEPTION = 'exc'
CLASS = 'class'
FILE = 'file'
LINE = 'line'
TRACE_ID = 'traceId'
CORR_ID = 'corId'
HTTP_CORID = 'HTTP_CORID'
RESPONSE_CODE = 'resCode'
RESPONSE_MESSAGE = 'resMsg'
HASH_SEPARATED = '<0..n>'
SKIP_INSTRUMENT = '<0:..>'
SPEAK = 'speak'
PAUSE = 'pause'
CONTINUE = 'continue'
CONFIG = 'config'
MAP = 'map'
REQUIRE = 'require'
COMMAND_CONN = 'commandConn'
DATA_CONN = 'dataConn'
ARF = 'arf'
FUNCTION_CALL = 'fnC'
FUNCTION_RET = 'fnR'
ARF_VER = 1
METRIC = 'm'
LOG = 'log'
PYTHON = 'python'
CONTAINER_ID = 'containerId'
PYTHON_MODULE = 'Python Module'
PYTHON_RASA_INTENT = 'Python Rasa Intent'
CORIDTEMPLATE = 'AgAAASVIQgAAAAFGAAAAAQAAABFqYXZhLnV0aWwuSGFzaE1hcAAAAARIQgAAAAJGAAAAAgAAABBqYXZhLmxhbmcuU3RyaW5nAApUeG5UcmFjZUlkSEIAAAADRQAAAAIAITZCOThGQURFMEE4NDk4MjU1MzI3MjFGRDhFNEE2M0RGM0hCAAAABEUAAAACABNDYWxsZXIgQ29tcG9uZW50IElESEIAAAAFRQAAAAIAATFIQgAAAAZFAAAAAgAPQ2FsbGVyVGltZXN0YW1wSEIAAAAHRQAAAAIADTE1NjA4NzgzNzE5NjFIQgAAAAhFAAAAAgARVXBzdHJlYW1HVUlEQ2FjaGVIQgAAAAlGAAAAAwAAABNqYXZhLnV0aWwuQXJyYXlMaXN0AAAAAA=='



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
import os
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.utils import network
from ca_apm_agent.python_modifiers.singleton import Singleton

logger = logging.getLogger(LOGGER_NAME[0] + '.probes.django.dbapi')

# Constants
SQLITE = 'sqlite'
SQLITE_STRING = 'SQLite DB'
FIREBIRD = 'firebird'
FIREBIRD_STRING = 'Firebird DB'
MYSQL = 'mysql'
MYSQL_STRING = 'MySQL DB'
POSTGRESQL = 'postgresql'
POSTGRESQL_STRING = 'PostgreSQL DB'
ORACLE = 'oracle'
ORACLE_STRING = 'Oracle DB'
SAP_SQL = 'sqlanywhere'
SAP_SQL_STRING = 'SAP SQL Anywhere DB'
MICROSOFT = 'microsoft'
MICROSOFT_SQLSERVER = 'sqlserver_ado'
MICROSOFT_ODBC = 'django_pyodbc'
MICROSOFT_SQLSERVER_STRING = 'MS SQL Server DB'
MICROSOFT_ODBC_STRING = 'MS ODBC'
HOST = 'HOST'
PORT = 'PORT'
NAME = 'NAME'
ENGINE = 'ENGINE'
QUERY = 'query'
DATABASE = 'database'
DATABASE_NAME = 'dbname'
URL = 'url'
COMMAND_TYPE = 'commandtype'
SQL = 'sql'
REGULAR = 'Regular'


class DBAPIProbe(Singleton):
    def __init__(self):
        self.host = network.getfqdn()

    def start(self, context):
        logger.debug('Function Arguments: %s **** %s', str(context.args), str(context.kwargs))
        vendor = str(context.args[0].db.vendor)
        context.func_name = vendor + '.' + QUERY

        database = None
        url = None
        host = context.args[0].db.settings_dict[HOST]
        port = context.args[0].db.settings_dict[PORT]
        name = context.args[0].db.settings_dict[NAME]
        if host in ('localhost', '127.0.0.1', '', ' ',):
            host = self.host
        if port in ('', ' ', None,):
            port = 'local'

        if vendor in (SQLITE, FIREBIRD,):  # For file and memory based DB's
            t = os.path.split(name)[1]
            if name[0] == '/':
                name = name[1:]
            url = host + ':' + port + '/' + name
            name = t
            if vendor == SQLITE:
                database = SQLITE_STRING
            elif vendor == FIREBIRD:
                database = FIREBIRD_STRING
            else:
                pass

        elif vendor in (MYSQL, POSTGRESQL, ORACLE, SAP_SQL, MICROSOFT,):  # For server based DB's
            url = host + ':' + port + '/' + name
            if vendor == MYSQL:
                database = MYSQL_STRING
            elif vendor == POSTGRESQL:
                database = POSTGRESQL_STRING
            elif vendor == ORACLE:
                database = ORACLE_STRING
            elif vendor == SAP_SQL:
                database = SAP_SQL_STRING
            elif vendor == MICROSOFT:
                engine = context.args[0].db.settings_dict[ENGINE].split('.')[-1]
                if engine == MICROSOFT_SQLSERVER:
                    database = MICROSOFT_SQLSERVER_STRING
                elif engine == MICROSOFT_ODBC:
                    database = MICROSOFT_ODBC_STRING
                else:
                    pass
        else:
            pass
        context.set_params(
            {DATABASE: database, DATABASE_NAME: name, HOST.lower(): host, PORT.lower(): port, URL: url,
             COMMAND_TYPE: REGULAR, SQL: str(context.args[1]), })
        logger.debug('Pushed params: %s', str(context.params))

    def finish(self, context):
        pass



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
import threading
from ca_apm_agent.connections.constants import TRACE_ID
from ca_apm_agent.utils import network


class Context:
    def __init__(self, seq, function_name=None, trace_id=None, params_dict=None, function_id=None, corr_id=None, timestamp=None):
        processid = os.getpid()
        containerid = network.getcontainerid()
        
        if containerid:
           processid = containerid
           
        self.pid = processid
        self.tid = threading.get_ident()
        self.seq = seq
        self.func_name = function_name
        self.func_id = function_id
        self.trace_id = str(trace_id)
        self.params = params_dict
        self.args = None
        self.kwargs = None
        self.exc = None
        self.cor_id = corr_id
        self.timestamp = None
        if self.params is None:
            self.params = {TRACE_ID: self.trace_id}
        else:
            self.params.update({TRACE_ID: self.trace_id})

    def push_target_params(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def set_trace_id(self, trace_id):
        self.trace_id = str(trace_id)
        if self.params is None:
            self.params = {TRACE_ID: self.trace_id}
        else:
            self.params.update({TRACE_ID: self.trace_id})
        
    def set_timestamp(self, timestamp):
        self.timestamp = timestamp
       
    def set_params(self, params_dict):
        if self.params is None:
            self.params = params_dict
        else:
            self.params.update(params_dict)

    def reset_params(self, params_dict=None):
        self.params = params_dict

    def set_exception(self, exception_dict):
        if self.exc is not None:
            self.exc.update(exception_dict)
        else:
            self.exc = exception_dict



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

pwd = path.abspath(path.dirname(__file__))

with open(path.join(pwd, 'LICENSE.txt'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ca-apm-agent',

    # Using version number from git tag
    # use_scm_version={"root": "..", "relative_to": __file__},
    version_format='{tag}.dev{commitcount}',
    setup_requires=['setuptools-git-version'],

    description='Python Agent for CA APM',
    long_description='Install by running:\n\n ``sudo pip install ca-apm-agent``\n\n' +
                     'Visit `CA DocOps <http://docops.ca.com/rest/ca/product/latest/topic?hid=HID_Python&space=APMSAAS&language=&format=rendered>`_ ' +
                     'for additional details\n\n' + long_description,

    # The project's main homepage.
    url='http://docops.ca.com/rest/ca/product/latest/topic?hid=HID_Python&space=APMSAAS&language=&format=rendered',

    # Author details
    author='CA, Inc.',
    author_email='TeamAPMLeo@ca.com',

    # Choose your license
    license='See package description',
	
	# Supported Python ranges
	python_requires='>=3.0',	

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Keep updating supported frameworks here

        'Framework :: Django',
	'Framework :: Flask',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',

        # Pick your license as you wish (should match "license" above)
        'License :: Other/Proprietary License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # Project associated keywords - space separated
    keywords='ca-apm-agent application-performance-monitoring APM apm CA',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,
    package_data={
        'directives': ['directives/agent.profile, directives/instrument.pbd'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'ca-apm-runpy=ca_apm_agent:main',
        ],
    },
)



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
import errno
import getopt
import sys
from ca_apm_agent.global_constants import PRINT_PREFIX

environ = os.environ

#
# To add more options follow the strategy below
# #Short Options
# options with an equivalent long name are to be added to the start of the short_opt_string
# options without an equivalent long name are to be appended to the short_opt_string
# #Long Options
# options with an equivalent short name are to be added to the start of the long_opt_list
# options without an equivalent short name are to be appended to the long_opt_list
#
short_opt_string = 'ha:i:d'
long_opt_list = ['help', 'agent-config=', 'instrumentation-config=', ]


def parse_opts(optlist):
    from ca_apm_agent.global_constants import CA_APM_PYTHON_ACFG, CA_APM_PYTHON_ICFG
    for opt, args in optlist:
        if opt in ('-h', '--help'):
            print (extended_usage())
            return True
        # Add options with both names after this comment
        elif opt in ('-a', '--agent-config'):
            environ[CA_APM_PYTHON_ACFG] = os.path.realpath(os.path.expanduser(args))
        elif opt in ('-i', '--instrumentation-config'):
            environ[CA_APM_PYTHON_ICFG] = os.path.realpath(os.path.expanduser(args))
        # Add options with only short names after this comment
        elif opt == '-d':
            # TODO enable debugging
            pass
        # Add options with only long names after this comment
        else:
            print_usage()
            return True
    return False


def print_usage():
    print (PRINT_PREFIX + '\n' \
            'For default usage: ca-apm-runpy <your existing command>\n ' \
            'For advanced usage ca-apm-runpy -h/--help')


def extended_usage():
    print (PRINT_PREFIX + '\n' \
          'For default usage                   :  ca-apm-runpy <your existing command>\n ' \
          'Advanced usage                      :  ca-apm-runpy <options> <your existing command>\n ' \
          'Available options \n' \
          '-h/--help                           :  Gets you back to this message \n' \
          '-a/--agent-config <file>            :  Custom agent configuration file \n' \
          '-i/--instrumentation config <file>  :  Custom instrumentation configuration file \n' \
          '' \
          'See documentation for details \n')


def main():
    args = sys.argv[1:]
    optlist, target = getopt.getopt(args, short_opt_string, long_opt_list)

    # option parsing
    help_quit = parse_opts(optlist)

    # bootstrap environment and target loading
    if not help_quit:
        print (PRINT_PREFIX + 'Setting up Environment')
        import ca_apm_agent.bootstrap
        pythonpath = [os.path.dirname(ca_apm_agent.bootstrap.__file__)]

        from ca_apm_agent.global_constants import PYTHONPATH, PYTHONUNBUFFERED, PYTHONVERBOSE
        if PYTHONPATH in environ:
            pythonpath.append(environ[PYTHONPATH])
        environ[PYTHONPATH] = ':'.join(pythonpath)

        environ[PYTHONUNBUFFERED] = 'True'
        # environ[PYTHONVERBOSE] = 'True'

        try:
            print (PRINT_PREFIX + 'Attempting to Bootstrap and Run "' + ' '.join(args) + '"')
            os.execvpe(target[0], target, environ)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                print (PRINT_PREFIX + '%s: no such file or directory' % target[0])
            elif exc.errno == errno.EPERM:
                 print (PRINT_PREFIX + '%s: permission denied' % target[0])
            raise



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""


# This is python 2 and 3 compatible
class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        init = getattr(cls._instances[cls], 'init', None)
        if callable(init):
            init()
        return cls._instances[cls]


class Singleton(_Singleton('SingletonMeta', (object,), {})):
    pass



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import os
from logging.handlers import BaseRotatingHandler


class RotatingFileHandler(BaseRotatingHandler):
    """
    Handler for logging to a set of files, which switches from one file
    to the next when the current file reaches a certain size.
    """

    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=0):
        """
        Open the specified file and use it as the stream for logging.

        By default, the file grows indefinitely. You can specify particular
        values of maxBytes and backupCount to allow the file to rollover at
        a predetermined size.

        Rollover occurs whenever the current log file is nearly maxBytes in
        length. If backupCount is >= 1, the system will successively create
        new files with the same pathname as the base file, but with extensions
        ".1", ".2" etc. appended to it. For example, with a backupCount of 5
        and a base file name of "app.log", you would get "app.log",
        "app.log.1", "app.log.2", ... through to "app.log.5". The file being
        written to is always "app.log" - when it gets filled up, it is closed
        and renamed to "app.log.1", and if files "app.log.1", "app.log.2" etc.
        exist, then they are renamed to "app.log.2", "app.log.3" etc.
        respectively.

        If maxBytes is zero, rollover never occurs.
        """
        # If rotation/rollover is wanted, it doesn't make sense to use another
        # mode. If for example 'w' were specified, then if there were multiple
        # runs of the calling application, the logs from previous runs would be
        # lost if the 'w' is respected, because the log file would be truncated
        # on each run.
        if maxBytes > 0:
            mode = 'a'
        if not os.path.isdir(os.path.dirname(os.path.abspath(filename))):
            os.makedirs(os.path.dirname(os.path.abspath(filename)), 493)
        BaseRotatingHandler.__init__(self, filename, mode, encoding, delay)
        self.maxBytes = maxBytes
        self.backupCount = backupCount
        if not hasattr(self, 'delay'):
            self.delay = delay  # This takes care of delay not being an inherited attribute, prior to python v2.7.6

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                sfn = "%s.%d" % (self.baseFilename, i)
                dfn = "%s.%d" % (self.baseFilename, i + 1)
                if os.path.exists(sfn):
                    # print "%s -> %s" % (sfn, dfn)
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self.baseFilename + ".1"
            if os.path.exists(dfn):
                os.remove(dfn)
            # Issue 18940: A file may not have been created if delay is True.
            if os.path.exists(self.baseFilename):
                os.rename(self.baseFilename, dfn)
        if not self.delay:
            self.stream = self._open()

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if self.stream is None:  # delay was set...
            self.stream = self._open()
        if self.maxBytes > 0:  # are we rolling over?
            msg = "%s\n" % self.format(record)
            self.stream.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self.stream.tell() + len(msg) >= self.maxBytes:
                return 1
        return 0



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import logging
from ca_apm_agent.global_constants import LOGGER_NAME
from ca_apm_agent.python_modifiers.singleton import Singleton
from ca_apm_agent.connections.constants import TRACE_ID, CORR_ID, PYTHON_MODULE
from ca_apm_agent.correlation import correlation

logger = logging.getLogger(LOGGER_NAME[0] + '.probes.flask.http')
# Constants
URL = 'url'
HOST_NAME = 'hostName'
HOST_PORT = 'hostPort'

class HTTPFlaskProbe(Singleton):
    def __init__(self):
        pass

    def start(self, context):
        environ = context.args[1]
        logger.debug("Flask Context Parameters : %s", str(environ))
        url = environ.get('RAW_URI')
        hostName = environ.get('HTTP_HOST')
        port = environ.get('SERVER_PORT')
        http_corid = environ.get('HTTP_CORID')
        
        if not http_corid:
           corId = correlation.get_new_corid_header()
           traceId = correlation.get_txn_traceid(corId)
           http_corid = corId
           context.cor_id = corId
        else:
           corId = correlation.upd_corid_header(http_corid)
           traceId = correlation.get_txn_traceid(corId)
           context.cor_id = corId
           
        if url is None:
           logger.debug("Flask URL is None setting to Default")
           url = "Default"
        else:
           logger.debug("Valid Flask request URL")  

        context.set_params({URL: url, HOST_NAME: hostName, HOST_PORT: port, TRACE_ID: traceId, CORR_ID: http_corid, PYTHON_MODULE: "Flask"})

    def finish(self, context):
        pass



__copyright__ = """
* Copyright (c) 2018 CA. All rights reserved.
*
* This software and all information contained therein is confidential and proprietary and
* shall not be duplicated, used, disclosed or disseminated in any way except as authorized
* by the applicable license agreement, without the express written permission of CA. All
* authorized reproductions must be marked with this language.
*
* EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE EXTENT
* PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE WITHOUT WARRANTY
* OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
* MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL CA BE
* LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR
* INDIRECT, FROM THE USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST
* PROFITS, BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
* EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
"""
import socket
import time
import errno


class Connect:
    def __init__(self, host, port, logger):
        self.host = host
        self.port = port
        self.logger = logger
        self.sock = None

    def connect(self):
        self.logger.info('Attempt to connect with the Collector')
        try:
            self.sock = socket.create_connection((self.host, self.port))
        except Exception as e:
            self.logger.error('Unable to connect to the Collector - Check if Collector is Running')
            self.logger.debug('Exception that was raised %s', str(e))
            return False
        else:
            self.logger.info('Attempt succeeded')
            return True

    def send(self, msg):
        msg = msg + '\r\n'
        MSGLEN = len(msg)
        totalsent = 0
        sent = 0
        while totalsent < MSGLEN:
            try:
                sent = self.sock.send(msg[totalsent:].encode())
            except Exception as e:
                if sent == 0:
                    self.logger.error('SEND - Socket connection broken/closed on the other end')
                return False
            else:
                self.logger.debug('Sent Message: ' + msg[totalsent:])
                totalsent = totalsent + sent
        return True

    def receive(self, timeout=1):
        self.sock.setblocking(0)
        total_data = []
        begin = time.time()
        while True:
            # if you got some data, then break after wait sec
            if total_data and time.time() - begin > timeout:
                break
            # if you got no data at all, wait a little longer
            elif time.time() - begin > timeout * 2:
                break
            try:
                data = self.sock.recv(8192).strip()
                if data:
                    total_data.append(data.decode())
                    begin = time.time()                    
                else:
                    self.logger.error('RECEIVE - Socket closed/broken on the other end')
                    return False
            except socket.error as e:
                if e.errno != errno.EWOULDBLOCK and e.errno != errno.EAGAIN:
                    self.logger.error('RECEIVE - Unknown error at socket')
                    self.logger.debug('Exception that was raised %s', str(e))
                    return False
        self.logger.debug('Received Message: ' + ''.join(str(total_data)))
        return ''.join(map(str, total_data))

    def close(self):
        try:
            self.sock.close()
        except AttributeError:
            pass
        except Exception as e:
            self.logger.error('Unknown exception closing socket')
            self.logger.debug('Exception that was raised %s', str(e))
